from cosmic_ray.operators import zero_iteration_loop
from cosmic_ray.work_record import WorkRecord
from cosmic_ray.worker import worker, WorkerOutcome

from path_utils import DATA_DIR, excursion, extend_path


def test_no_test_return_value():
    with extend_path(DATA_DIR), excursion(DATA_DIR):
        result = worker("a.b", zero_iteration_loop.ZeroIterationLoop,
                        100, "pytest")
        expected = WorkRecord(
            data=None,
            test_outcome=None,
            worker_outcome=WorkerOutcome.NO_TEST,
            diff=None,
            module=None,
            operator=None,
            occurrence=None,
            line_number=None,
            command_line=None,
            job_id=None)

        assert result == expected



"""Tests for the various mutation operators.
"""
import ast
import copy
import pytest

from cosmic_ray.operators.comparison_operator_replacement import \
    MutateComparisonOperator
from cosmic_ray.operators.unary_operator_replacement import \
    MutateUnaryOperator
from cosmic_ray.operators.binary_operator_replacement import \
    MutateBinaryOperator
from cosmic_ray.counting import _CountingCore
from cosmic_ray.operators.boolean_replacer import (ReplaceTrueFalse,
                                                   ReplaceAndWithOr,
                                                   ReplaceOrWithAnd,
                                                   AddNot)
from cosmic_ray.operators.break_continue import (ReplaceBreakWithContinue,
                                                 ReplaceContinueWithBreak)
from cosmic_ray.operators.exception_replacer import ExceptionReplacer
from cosmic_ray.operators.number_replacer import NumberReplacer
from cosmic_ray.operators.remove_decorator import RemoveDecorator
from cosmic_ray.operators.zero_iteration_loop import ZeroIterationLoop
from cosmic_ray.mutating import MutatingCore


class Linearizer(ast.NodeVisitor):
    """A NodeVisitor which builds a linear list of nodes it visits.

    The basic point is to be able to take a tree of nodes and reproducably
    construct a simple list. This list is useful for e.g. comparison between
    trees.

    After using this to visit an AST, the `nodes` attribute holds the list of
    nodes.
    """
    def __init__(self):
        self.nodes = []

    def generic_visit(self, node):
        self.nodes.append(node)
        super().generic_visit(node)


def linearize_tree(node):
    "Given an AST, return a list of the nodes therein."
    l = Linearizer()
    l.visit(node)
    return l.nodes


OPERATOR_SAMPLES = [
    (ReplaceTrueFalse, 'True'),
    (ReplaceAndWithOr, 'if True and False: pass'),
    (ReplaceOrWithAnd, 'if True or False: pass'),
    (AddNot, 'if True or False: pass'),
    (AddNot, 'A if B else C'),
    (AddNot, 'assert isinstance(node, ast.Break)'),
    (AddNot, 'while True: pass'),
    (ReplaceBreakWithContinue, 'while True: break'),
    (ReplaceContinueWithBreak, 'while False: continue'),
    (NumberReplacer, 'x = 1'),
    (MutateComparisonOperator, 'if x > y: pass'),
    (MutateUnaryOperator, 'return not X'),
    (MutateUnaryOperator, 'x = -1'),
    (MutateBinaryOperator, 'x * y'),
    (MutateBinaryOperator, 'x - y'),
    (ExceptionReplacer, 'try: raise OSError \nexcept OSError: pass'),
    (ZeroIterationLoop, 'for i in range(1,2): pass'),
    (RemoveDecorator, 'def wrapper(f): f.cosmic_ray=1; return f\n@wrapper\ndef foo(): pass')
]


@pytest.mark.parametrize('operator,code', OPERATOR_SAMPLES)
def test_activation_record_created(operator, code):
    node = ast.parse(code)
    core = MutatingCore(0)
    op = operator(core)
    assert core.activation_record is None
    op.visit(node)
    assert core.activation_record is not None


@pytest.mark.parametrize('operator,code', OPERATOR_SAMPLES)
def test_no_activation_record_created(operator, code):
    node = ast.parse(code)
    core = MutatingCore(-1)
    op = operator(core)
    op.visit(node)
    assert core.activation_record is None


@pytest.mark.parametrize('operator,code', OPERATOR_SAMPLES)
def test_mutation_changes_ast(operator, code):
    node = ast.parse(code)
    core = MutatingCore(0)
    mutant = operator(core).visit(copy.deepcopy(node))

    orig_nodes = linearize_tree(node)
    mutant_nodes = linearize_tree(mutant)

#todo: disabled b/c adding/removing the not keyword
# changes the number of nodes in the tree.
#    assert len(orig_nodes) == len(mutant_nodes)

    assert ast.dump(node) != ast.dump(mutant)


@pytest.mark.parametrize('operator,code', OPERATOR_SAMPLES)
def test_no_mutation_leaves_ast_unchanged(operator, code):
    node = ast.parse(code)

    core = MutatingCore(-1)
    replacer = operator(core)
    assert ast.dump(node) == ast.dump(replacer.visit(copy.deepcopy(node)))


@pytest.mark.parametrize('operator,code', OPERATOR_SAMPLES)
def test_replacement_activated_core(operator, code):
    node = ast.parse(code)
    core = _CountingCore()
    op = operator(core)
    op.visit(node)
    assert core.count >= 1



def make_record(name, fields=(), docstring=""):
    """Create a new record class.

    A Record is fundamentally a dict with a specified set of keys. These keys
    will always have a value (defaulting to None), they can't be removed, and
    new keys can not be added.

    This may sound a lot like a class, and that's true. The main benefit of
    records is that they can be treated directly like dicts for the most part,
    and, critically, they are easy to JSON-ify. Also, like classes, they ensure
    that they're only used in the correct way, i.e. users can only access the
    specified fields. This prevents the confusion of using simple dicts where
    people can use conflicting or confusing key names.

    Args:
        name: The name of the class to be created.
        fields: The names of the fields in the record.
        docstring: The docstring for the record class.

    Returns: A new class derived from dict with the specified fields.

    """
    def __init__(self, vals=None, **kwargs):
        dict.__init__(self, dict.fromkeys(fields))
        vals = vals or dict()
        kwargs.update(vals)
        for k, v in kwargs.items():
            self[k] = v

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError('no attribute {}'.format(name))

    def __setattr__(self, name, value):
        try:
            self[name] = value
        except KeyError:
            raise AttributeError('no attribute {}'.format(name))

    def __getitem__(self, name):
        if name not in self:
            raise KeyError('no field {} in record'.format(name))
        return dict.__getitem__(self, name)

    def __setitem__(self, name, value):
        if name not in self:
            raise KeyError('no field {} in record'.format(name))
        dict.__setitem__(self, name, value)

    def __delitem__(self, name):
        raise KeyError('record does not support deleting fields: {}'.format(name))

    def update(self, d):
        for k, v in d.items():
            self[k] = v

    attrs = {
        '__init__': __init__,
        '__getattr__': __getattr__,
        '__setattr__': __setattr__,
        '__getitem__': __getitem__,
        '__setitem__': __setitem__,
        '__delitem__': __delitem__,
        'update': update
    }

    rec = type(name, (dict,), attrs)
    rec.__doc__ = docstring
    return rec


WorkRecord = make_record(
    'WorkRecord',

    [
        # Arbitrary data returned by the concrete TestRunner to provide more
        # information about the test results.
        'data',

        # A test_runner.TestOutcome from the test run.
        'test_outcome',

        # A worker.WorkOutcome describing how the worker completed.
        'worker_outcome',

        # The diff produced by the operators
        'diff',

        # the module to be mutated
        'module',

        # The name of the operators
        'operator',

        # The occurrence on which the operator was applied.
        'occurrence',

        # The line number at which the operator was applied.
        'line_number',

        'command_line',
        'job_id'
    ],
    docstring=" The details of a specific mutation and test run in CosmicRay."
    )



import os
import nose

from .test_runner import TestRunner
from cosmic_ray.util import redirect_stdout, redirect_stderr


class NoseResultsCollector(nose.plugins.Plugin):
    name = 'cosmic_ray'
    enabled = True

    def __init__(self):
        super(self.__class__, self).__init__()
        self.result = None

    def finalize(self, result):
        self.result = result


class NoseRunner(TestRunner):
    """A TestRunner using nosetest.

    This treats `test_args` as a list of arguments to `nose.run()`. The args
    are passed directly to that function. See nose's command line reference
    for a description of what arguments are accepted.

    NOTE: ``-s`` is not accepted here!
    """

    def _run(self):
        argv = ['', '--with-cosmic_ray']
        argv += list(self.test_args)
        collector = NoseResultsCollector()

        with open(os.devnull, 'w') as devnull, \
            redirect_stdout(devnull), redirect_stderr(devnull):
            nose.run(argv=argv, plugins=[collector])
        return (collector.result.wasSuccessful(),
                [r[1] for r in collector.result.errors +
                               collector.result.failures])



import ast

from cosmic_ray.util import get_line_number


def test_basic():
    code = 'if x == 1:' \
           '    return 3'
    tree = ast.parse(code)

    assert get_line_number(tree) == '<UNKNOWN>'
    assert get_line_number(tree.body[0]) == 1
    assert get_line_number(tree.body[0].test) == 1
    assert get_line_number(tree.body[0].test.ops[0]) == '<UNKNOWN>'



import ast
import builtins

from .operator import Operator


class CosmicRayTestingException(Exception):
    pass


setattr(builtins,
        CosmicRayTestingException.__name__,
        CosmicRayTestingException)


class ExceptionReplacer(Operator):
    """An operator that modifies exception handlers."""

    def visit_ExceptHandler(self, node):  # noqa
        return self.visit_mutation_site(node)

    def mutate(self, node, _):
        """Modify the exception handler with another exception type."""
        except_id = CosmicRayTestingException.__name__
        except_type = ast.Name(id=except_id, ctx=ast.Load())
        new_node = ast.ExceptHandler(type=except_type, name=node.name,
                                     body=node.body)
        return new_node



# Database that keeps track of mutation testing work progress.
#
# Essentially, there's a row in the DB for each mutation that needs to be
# executed in some run. These initially start off with no results, and results
# are added as they're completed.
#
# What describes a work-item?
#  - module name
#  - operator name
#  - occurrence
#  - test runner name
#  - test args
#  - timeout
#
# What describes results?
#  - one of the strings 'exception', 'no-test', or 'normal'
#  - for 'exception', the sys.exc_info() of the exception
#  - for 'no-test', there are no other results
#  - for 'normal', there is an *activation-record* (a dict) and a
#    test_runner.TestResult.

import contextlib
from enum import Enum
import os

# This db may well not scale very well. We need to be ready to switch it out
# for something quicker if not. But for now it's *very* convenient.
import tinydb

from .work_record import WorkRecord


class WorkDB:
    class Mode(Enum):
        # Open existing files, creating if necessary
        create = 1,

        # Open only existing files, failing if it doesn't exist
        open = 2

    def __init__(self, path, mode):
        """Open a DB in file `path` in mode `mode`.

        Args:
          path: The path to the DB file.
          mode: The mode in which to open the DB. See the `Mode` enum for
            details.

        Raises:
          FileNotFoundError: If `mode` is `Mode.open` and `path` does not
            exist.
        """
        if (mode == WorkDB.Mode.open) and (not os.path.exists(path)):
            raise FileNotFoundError(
                'Requested file {} not found'.format(path))

        self._db = tinydb.TinyDB(path)

    def close(self):
        self._db.close()

    @property
    def _work_parameters(self):
        """The table of work parameters."""
        return self._db.table('work-parameters')

    @property
    def _work_items(self):
        """The table of work items."""
        return self._db.table('work-items')

    def set_work_parameters(self, test_runner, test_args, timeout):
        """Set (replace) the work parameters for the session.

        Args:
          test_runner: The name of the test runner plugin to use.
          test_args: The arguments to pass to the test runner.
          timeout: The timeout for tests.
        """
        table = self._work_parameters
        table.purge()
        table.insert({
            'test-runner': test_runner,
            'test-args': test_args,
            'timeout': timeout,
        })

    def get_work_parameters(self):
        """Get the work parameters (if set) for the session.

        Returns: a tuple of `(test-runner, test-args, timeout)`.

        Raises:
          ValueError: If there are no work parameters set for the session.
        """
        table = self._work_parameters

        try:
            record = table.all()[0]
        except IndexError:
            raise ValueError('work-db has no work parameters')

        return (record['test-runner'],
                record['test-args'],
                record['timeout'])

    def add_work_records(self, records):
        """Add a sequence of WorkRecords.

        Args:
          items: An iterable of tuples of the form `(module-name,
            operator-name, occurrence)`.

        """
        self._work_items.insert_multiple(records)

    def clear_work_records(self):
        """Clear all work items from the session.

        This removes any associated results as well.
        """
        self._work_items.purge()

    @property
    def work_records(self):
        """The sequence of `WorkItem`s in the session.

        This include both complete and incomplete items.

        Each work item is a dict with the keys `module-name`, `op-name`, and
        `occurrence`. Items with results will also have the keys `results-type`
        and `results-data`.

        """
        return (WorkRecord(r) for r in self._work_items.all())

    def update_work_record(self, work_record):
        """Updates an existing WorkRecord by job_id.

        Args:
            work_record: A WorkRecord representing the new state of a job.

        Raises:
          KeyError: If there is no existing record with the same job_id.
        """
        self._work_items.update(
            work_record,
            tinydb.Query().job_id == work_record.job_id)

    @property
    def pending_work(self):
        """The sequence of pending `WorkItem`s in the session."""
        table = self._work_items
        work_item = tinydb.Query()
        pending = table.search(work_item.worker_outcome == None)
        return (WorkRecord(r) for r in pending)


@contextlib.contextmanager
def use_db(path, mode=WorkDB.Mode.create):
    """
    Open a DB in file `path` in mode `mode` as a context manager.

    On exiting the context the DB will be automatically closed.

    Args:
      path: The path to the DB file.
      mode: The mode in which to open the DB. See the `Mode` enum for
        details.

    Raises:
      FileNotFoundError: If `mode` is `Mode.open` and `path` does not
        exist.
    """
    db = WorkDB(path, mode)
    try:
        yield db
    except Exception:
        db.close()
        raise



import ast
import sys
from .operator import Operator


class ReplaceTrueFalse(Operator):
    """An operator that modifies True/False constants."""
    def visit_NameConstant(self, node):  # noqa
        """
            New in version 3.4: Previously, these constants were instances of ``Name``.
            http://greentreesnakes.readthedocs.io/en/latest/nodes.html#NameConstant
        """
        if node.value in [True, False]:
            return self.visit_mutation_site(node)
        else:
            return node

    def visit_Name(self, node):  # noqa
        """For backward compatibility with Python 3.3."""
        if node.id in ['True', 'False']:
            return self.visit_mutation_site(node)
        else:
            return node

    def mutate(self, node, _):
        """Modify the boolean value on `node`."""
        if sys.version_info >= (3, 4):
            return ast.NameConstant(value=not node.value)
        else:
            return ast.Name(id=not ast.literal_eval(node.id), ctx=node.ctx)


class ReplaceAndWithOr(Operator):
    """An operator that swaps 'and' with 'or'."""
    def visit_BoolOp(self, node):  # noqa
        """
            http://greentreesnakes.readthedocs.io/en/latest/nodes.html#BoolOp
        """
        if isinstance(node.op, ast.And):
            return self.visit_mutation_site(node, len(node.values))
        else:
            return node

    def mutate(self, node, idx):
        """Replace AND with OR."""
        # replace all occurences of And()
        # A and B and C -> A or B or C
        node.op = ast.Or()

        # or replace an operator somewhere in the middle
        # of the expression
        if idx and len(node.values) > 2:
            left = node.values[:idx]
            if len(left) > 1:
                left = [ast.BoolOp(op=ast.And(), values=left)]

            right = node.values[idx:]
            if len(right) > 1:
                right = [ast.BoolOp(op=ast.And(), values=right)]

            node.values = []
            node.values.extend(left)
            node.values.extend(right)

        return node


class ReplaceOrWithAnd(Operator):
    """An operator that swaps 'or' with 'and'."""
    def visit_BoolOp(self, node):  # noqa
        """
            http://greentreesnakes.readthedocs.io/en/latest/nodes.html#BoolOp
        """
        if isinstance(node.op, ast.Or):
            return self.visit_mutation_site(node, len(node.values))
        else:
            return node

    def mutate(self, node, idx):
        """Replace OR with AND."""
        if idx and len(node.values) > 2:
            left_list = node.values[:idx-1]
            right_list = node.values[idx+1:]
            left = node.values[idx-1]
            right = node.values[idx]

            new_node = ast.BoolOp(op=ast.And(), values=[left, right])

            node.values = []
            node.values.extend(left_list)
            node.values.append(new_node)
            node.values.extend(right_list)
        else:
            node.op = ast.And()

        return node


class AddNot(Operator):

    """
        An operator that adds the 'not' keyword to boolean expressions.

        NOTE: 'not' as unary operator is mutated in `unary_operator_replacement.py`,
        including deletion of the same operator.
    """

    def visit_If(self, node):  # noqa
        return self.visit_mutation_site(node)

    def visit_IfExp(self, node):  # noqa
        return self.visit_mutation_site(node)

    def visit_Assert(self, node):  # noqa
        return self.visit_mutation_site(node)

    def visit_While(self, node):  # noqa
        return self.visit_mutation_site(node)

    def mutate(self, node, _):
        """
        Add the 'not' keyword.

        Note: this will negate the entire if condition.
        """
        if hasattr(node, 'test'):
            node.test = ast.UnaryOp(op=ast.Not(), operand=node.test)
            return node



"""Functions related to finding modules for testing."""

import importlib
import logging
import os
import pkgutil
import re

LOG = logging.getLogger()


def fixup_module_name(name):
    """If `name` appears to be a source file name, this converts it to a module
    name. Otherwise this returns `name` inchanged.

    For example, if you pass in "foo/bar.py", this will return "foo.bar". If
    you pass in "llama.yak", this just returns "llama.yak".

    This is primarily a convenience function to allow users to use
    bash-completion when specifying module names.
    """
    if os.path.exists(name):
        if name.endswith('.py'):
            name = name[:-3]
        name = name.replace('/', '.')
    return name


def find_modules(name, excludes=None):
    """Generate sequence of all submodules of NAME, including NAME itself.

    `excludes` is a sequence of regular expression. If a full module name
    matches any of these expressions then it won't be returned in the results.
    Critically, no module beneath an excluded module will be reported either.

    Given a directory structure like this:

        /a/
          __init__.py
          b.py
          c/
              __init__.py
              d.py

    you get this:

        >>> list(find_modules('a'))
        [<module 'a' from 'a/__init__.py'>,
         <module 'a.b' from 'a/b.py'>,
         <module 'a.c' from 'a/c/__init__.py'>,
         <module 'a.c.d' from 'a/c/d.py'>]

    """
    excludes = excludes or []
    module_names = [name]
    exclude_patterns = [re.compile(ex) for ex in excludes]
    while module_names:
        module_name = module_names.pop()
        try:

            if any([pattern.match(module_name)
                    for pattern in exclude_patterns]):
                continue

            module = importlib.import_module(module_name)

            yield module

            if hasattr(module, '__path__'):
                for _, name, _ in pkgutil.iter_modules(module.__path__):
                    module_names.append(
                        '{}.{}'.format(
                            module_name, name))
        except Exception:  # pylint:disable=broad-except
            LOG.exception(
                'Unable to import %s',
                module_name)



import os
import pytest

from .test_runner import TestRunner
from cosmic_ray.util import redirect_stdout


class ResultCollector:
    def __init__(self):
        self.reports = []

    def pytest_runtest_logreport(self, report):
        self.reports.append(report)


class PytestRunner(TestRunner):
    """A TestRunner using pytest.

    This treats `test_args` as a list of arguments to `pytest.main()`. The args
    are passed directly to that function, so see it's documentation for a
    description of how the arguments are used.

    """

    def _run(self):
        collector = ResultCollector()

        with open(os.devnull, 'w') as devnull, redirect_stdout(devnull):
            pytest.main(list(self.test_args),
                        plugins=[collector])

        return (
            all(not r.failed for r in collector.reports),
            [(repr(r), r.longreprtext) for r in collector.reports if r.failed])



from cosmic_ray.testing.test_runner import TestOutcome
from cosmic_ray.work_record import WorkRecord
from cosmic_ray.worker import WorkerOutcome
import docopt
import json
import sys
import xml.etree.ElementTree


def _print_item(work_record, full_report):
    data = work_record.data
    outcome = work_record.worker_outcome
    if outcome in [WorkerOutcome.NORMAL, WorkerOutcome.EXCEPTION]:
        outcome = work_record.test_outcome
    ret_val = [
        'job ID {}:{}:{}'.format(
            work_record.job_id,
            outcome,
            work_record.module),
        'command: {}'.format(
            ' '.join(work_record.command_line)
            if work_record.command_line is not None else ''),
    ]
    if outcome == TestOutcome.KILLED and not full_report:
        ret_val = []
    elif work_record.worker_outcome == WorkerOutcome.TIMEOUT:
        if full_report:
            ret_val.append("timeout: {:.3f} sec".format(data))
        else:
            ret_val = []
    elif work_record.worker_outcome in [WorkerOutcome.NORMAL, WorkerOutcome.EXCEPTION]:
        ret_val += data
        ret_val += work_record.diff

    # for presentation purposes only
    if ret_val:
        ret_val.append('')

    return ret_val


def is_killed(record):
    if record.worker_outcome == WorkerOutcome.TIMEOUT:
        return True
    elif record.worker_outcome == WorkerOutcome.NORMAL:
        if record.test_outcome == TestOutcome.KILLED:
            return True
    return False


def create_report(records, show_pending, full_report=False):
    total_jobs = 0
    pending_jobs = 0
    kills = 0
    for item in records:
        total_jobs += 1
        if item.worker_outcome is None:
            pending_jobs += 1
        if is_killed(item):
            kills += 1
        if (item.worker_outcome is not None) or show_pending:
            yield from _print_item(item, full_report)

    completed_jobs = total_jobs - pending_jobs

    yield 'total jobs: {}'.format(total_jobs)

    if completed_jobs > 0:
        yield 'complete: {} ({:.2f}%)'.format(
            completed_jobs, completed_jobs / total_jobs * 100)
        yield 'survival rate: {:.2f}%'.format(
            (1 - kills / completed_jobs) * 100)
    else:
        yield 'no jobs completed'


def survival_rate():
    """cr-rate

Usage: cr-rate

Read JSON work-records from stdin and print the survival rate.
"""
    records = (WorkRecord(json.loads(line)) for line in sys.stdin)

    total_jobs = 0
    pending_jobs = 0
    kills = 0
    for item in records:
        total_jobs += 1
        if item.worker_outcome is None:
            pending_jobs += 1
        if is_killed(item):
            kills += 1

    completed_jobs = total_jobs - pending_jobs

    if not completed_jobs:
        rate = 0
    else:
        rate = (1 - kills / completed_jobs) * 100

    print('{:.2f}'.format(rate))


def report():
    """cr-report

Usage: cr-report [--full-report] [--show-pending]

Print a nicely formatted report of test results and some basic statistics.

options:
    --full-report   Show test output and mutation diff for killed mutants
    --show-pending  Display results for incomplete tasks
"""

    arguments = docopt.docopt(report.__doc__, version='cr-format 0.1')
    full_report = arguments['--full-report']
    show_pending = arguments['--show-pending']
    records = (WorkRecord(json.loads(line)) for line in sys.stdin)
    for line in create_report(records, show_pending, full_report):
        print(line)


def _create_element_from_item(work_record):
    data = work_record.data
    sub_elem = xml.etree.ElementTree.Element('testcase')

    sub_elem.set('classname', work_record.job_id)
    sub_elem.set('line', str(work_record.line_number))
    sub_elem.set('file', work_record.module)
    if work_record.command_line:
        sub_elem.set('name', str(work_record.command_line))

    outcome = work_record.worker_outcome

    if outcome == WorkerOutcome.TIMEOUT:
        error_elem = xml.etree.ElementTree.SubElement(sub_elem, 'error')
        error_elem.set('message', "Timeout: {:.3f} sec".format(data))
    elif outcome == WorkerOutcome.EXCEPTION:
        error_elem = xml.etree.ElementTree.SubElement(sub_elem, 'error')
        error_elem.set('message', "Worker has encountered exception")
        error_elem.text = str(data) + "\n".join(work_record.diff)
    elif (outcome == WorkerOutcome.NORMAL and
          work_record.test_outcome in [TestOutcome.SURVIVED, TestOutcome.INCOMPETENT]):
        failure_elem = xml.etree.ElementTree.SubElement(sub_elem, 'failure')
        failure_elem.set('message', "Mutant has survived your unit tests")
        failure_elem.text = str(data) + "\n".join(work_record.diff)

    return sub_elem


def _create_xml_report(records):
    total_jobs = 0
    errors = 0
    failed = 0
    root_elem = xml.etree.ElementTree.Element('testsuite')
    for item in records:
        total_jobs += 1
        if item.worker_outcome is None:
            errors += 1
        if is_killed(item):
            failed += 1
        if item.worker_outcome is not None:
            subelement = _create_element_from_item(item)
            root_elem.append(subelement)

    root_elem.set('errors', str(errors))
    root_elem.set('failures', str(failed))
    root_elem.set('skips', str(0))
    root_elem.set('tests', str(total_jobs))
    return xml.etree.ElementTree.ElementTree(root_elem)


def report_xml():
    """cr-xml

Usage: cr-xml

Print an XML formatted report of test results for continuos integration systems
"""
    records = (WorkRecord(json.loads(line)) for line in sys.stdin)
    xml_elem = _create_xml_report(records)
    xml_elem.write(sys.stdout.buffer, encoding='utf-8', xml_declaration=True)



from stevedore import driver, ExtensionManager


def get_operator(name):
    """Get an operator class from a plugin.

    Attrs:
        name: The name of the plugin containing the operator class.

    Returns: The operator *class object* (i.e. not an instance) provided by the
        plugin named `name`.
    """
    return ExtensionManager('cosmic_ray.operators')[name].plugin


def operator_names():
    """Get an iterable of all operator plugin names."""
    return ExtensionManager('cosmic_ray.operators').names()


def get_test_runner(name, test_args):
    """Get a test-runner instance by name."""
    test_runner_manager = driver.DriverManager(
        namespace='cosmic_ray.test_runners',
        name=name,
        invoke_on_load=True,
        invoke_args=(test_args,),
    )

    return test_runner_manager.driver


def test_runner_names():
    """Get iterable of test-runner plugin names."""
    return ExtensionManager('cosmic_ray.test_runners').names()



"""This module contains mutation operators which replace one
comparison operator with another.
"""

import ast
from .operator import Operator
from ..util import build_mutations


OPERATORS = (ast.Eq, ast.NotEq, ast.Lt, ast.LtE, ast.Gt, ast.GtE,
             ast.Is, ast.IsNot, ast.In, ast.NotIn)


def _to_ops(from_op):
    """The sequence of operators which `from_op` could be mutated to.

    There are a number of potential replacements which we avoid because they
    almost always produce equivalent mutants. We encode those in the logic
    here.

    NB: This is an imperfect, stop-gap solution to the problem of certain
    equivalent mutants. Obviously `==` is not generally the same as `is`, but
    that mutation is also a source of a good number of equivalents. The real
    solution to this problem will probably come in the form of real exception
    declarations or something.

    See https://github.com/sixty-north/cosmic-ray/pull/162 for some more
    discussion of this issue.

    """

    for to_op in OPERATORS:
        if isinstance(from_op, ast.Eq) and to_op is ast.Is:
            pass
        elif isinstance(from_op, ast.NotEq) and to_op is ast.IsNot:
            pass
        else:
            yield to_op


class MutateComparisonOperator(Operator):
    """An operator that modifies comparisons."""

    def visit_Compare(self, node):
        """
            http://greentreesnakes.readthedocs.io/en/latest/nodes.html#Compare
        """
        return self.visit_mutation_site(
            node,
            len(build_mutations(node.ops, _to_ops)))

    def mutate(self, node, idx):
        from_idx, to_op = build_mutations(node.ops, _to_ops)[idx]
        node.ops[from_idx] = to_op()
        return node



"""This is the command-line program for cosmic ray.

Here we manage command-line parsing and launching of the internal
machinery that does mutation testing.
"""
import itertools
import json
import logging
import os
import pprint
import subprocess
import sys

import docopt_subcommands as dsc
import transducer.eager
from transducer.functional import compose
import transducer.lazy
from transducer.transducers import filtering, mapping

import cosmic_ray.commands
import cosmic_ray.counting
import cosmic_ray.modules
import cosmic_ray.worker
from cosmic_ray.testing.test_runner import TestOutcome
from cosmic_ray.timing import Timer
from cosmic_ray.util import redirect_stdout
from cosmic_ray.work_db import use_db, WorkDB


LOG = logging.getLogger()

REMOVE_COMMENTS = mapping(lambda x: x.split('#')[0])
REMOVE_WHITESPACE = mapping(str.strip)
NON_EMPTY = filtering(bool)
CONFIG_FILE_PARSER = compose(REMOVE_COMMENTS,
                             REMOVE_WHITESPACE,
                             NON_EMPTY)


def _load_file(config_file):
    """Read configuration from a file.

    This reads `config_file`, yielding each non-empty line with
    whitespace and comments stripped off.
    """
    with open(config_file, 'rt', encoding='utf-8') as f:
        yield from transducer.lazy.transduce(CONFIG_FILE_PARSER, f)


@dsc.command()
def handle_load(config):
    """usage: {program} load <config-file>

Load a command configuration from <config-file> and run it.

A "command configuration" is simply a command-line invocation for cosmic-ray,
where each token of the command is on a separate line.
    """
    filename = config['<config-file>']
    argv = _load_file(filename)
    return main(argv=list(argv))


@dsc.command()
def handle_baseline(configuration):
    """usage: {program} baseline [options] <top-module> [-- <test-args> ...]

Run an un-mutated baseline of <top-module> using the tests in <test-dir>.
This is largely like running a "worker" process, with the difference
that a baseline run doesn't mutate the code.

options:
  --no-local-import   Allow importing module from the current directory
  --test-runner=R     Test-runner plugin to use [default: unittest]
"""
    sys.path.insert(0, '')
    test_runner = cosmic_ray.plugins.get_test_runner(
        configuration['--test-runner'],
        configuration['<test-args>']
    )

    work_record = test_runner()
    # note: test_runner() results are meant to represent
    # status codes when executed against mutants.
    # SURVIVED means that the test suite executed without any error
    # hence CR thinks the mutant survived. However when running the
    # baseline execution we don't have mutations and really want the
    # test suite to report PASS, hence the comparison below!
    if work_record.test_outcome != TestOutcome.SURVIVED:
        # baseline failed, print whatever was returned
        # from the test runner and exit
        LOG.error('baseline failed')
        print(''.join(work_record.data))
        sys.exit(2)


def _get_db_name(session_name):
    if session_name.endswith('.json'):
        return session_name
    else:
        return '{}.json'.format(session_name)


@dsc.command()
def handle_init(configuration):
    """usage: {program} init [options] [--exclude-modules=P ...] (--timeout=T | --baseline=M) <session-name> <top-module> [-- <test-args> ...]

Initialize a mutation testing run. The primarily creates a database of "work to
be done" which describes all of the mutations and test runs that need to be
executed for a full mutation testing run. The testing run will mutate
<top-module> (and submodules) using the tests in <test-dir>. This doesn't
actually run any tests. Instead, it scans the modules-under-test and simply
generates the work order which can be executed with other commands.

The session-name argument identifies the run you're creating. Its most
important role is that it's used to name the database file.

options:
  --no-local-import   Allow importing module from the current directory
  --test-runner=R     Test-runner plugin to use [default: unittest]
  --exclude-modules=P Pattern of module names to exclude from mutation
    """
    # This lets us import modules from the current directory. Should probably
    # be optional, and needs to also be applied to workers!
    sys.path.insert(0, '')

    if configuration['--timeout'] is not None:
        timeout = float(configuration['--timeout'])
    else:
        baseline_mult = float(configuration['--baseline'])
        assert baseline_mult is not None
        command = 'cosmic-ray baseline --test-runner={test_runner} {module} -- {test_args}'.format(
            test_runner=configuration['--test-runner'],
            module=configuration['<top-module>'],
            test_args=' '.join(configuration['<test-args>'])
        )

        # We run the baseline in a subprocess to more closely emulate the
        # runtime of a worker subprocess.
        with Timer() as t:
            subprocess.check_call(command.split())

        timeout = baseline_mult * t.elapsed.total_seconds()

    LOG.info('timeout = %f seconds', timeout)

    modules = set(
        cosmic_ray.modules.find_modules(
            cosmic_ray.modules.fixup_module_name(configuration['<top-module>']),
            configuration['--exclude-modules']))

    LOG.info('Modules discovered: %s', [m.__name__ for m in modules])

    db_name = _get_db_name(configuration['<session-name>'])

    with use_db(db_name) as db:
        cosmic_ray.commands.init(
            modules,
            db,
            configuration['--test-runner'],
            configuration['<test-args>'],
            timeout)


@dsc.command()
def handle_exec(configuration):
    """usage: {program} exec [--dist] <session-name>

Perform the remaining work to be done in the specified session. This requires
that the rest of your mutation testing infrastructure (e.g. worker processes)
are already running.

options:
    --dist  Distribute tests to remote workers
    """
    db_name = _get_db_name(configuration['<session-name>'])
    dist = configuration['--dist']

    with use_db(db_name, mode=WorkDB.Mode.open) as db:
        cosmic_ray.commands.execute(db, dist)


@dsc.command()
def handle_run(configuration):
    """usage: {program} run [options] [--dist] [--exclude-modules=P ...] (--timeout=T | --baseline=M) <session-name> <top-module> [-- <test-args> ...]

This simply runs the "init" command followed by the "exec" command.

It's important to remember that "init" clears the session database, including
any results you may have already received. So DO NOT USE THIS COMMAND TO
CONTINUE EXECUTION OF AN INTERRUPTED SESSION! If you do this, you will lose any
existing results.

options:
  --no-local-import   Allow importing module from the current directory
  --test-runner=R     Test-runner plugin to use [default: unittest]
  --exclude-modules=P Pattern of module names to exclude from mutation

    """
    handle_init(configuration)
    handle_exec(configuration)


@dsc.command()
def handle_dump(configuration):
    """usage: {program} dump <session-name>

JSON dump of session data.
    """
    db_name = _get_db_name(configuration['<session-name>'])

    with use_db(db_name, WorkDB.Mode.open) as db:
        for record in db.work_records:
            print(json.dumps(record))


@dsc.command()
def handle_counts(configuration):
    """usage: {program} counts [options] [--exclude-modules=P ...] <top-module>

Count the number of tests that would be run for a given testing configuration.
This is mostly useful for estimating run times and keeping track of testing
statistics.

options:
  --no-local-import   Allow importing module from the current directory
  --test-runner=R     Test-runner plugin to use [default: unittest]
  --exclude-modules=P Pattern of module names to exclude from mutation
"""
    sys.path.insert(0, '')

    modules = cosmic_ray.modules.find_modules(
        cosmic_ray.modules.fixup_module_name(configuration['<top-module>']),
        configuration['--exclude-modules'])

    operators = cosmic_ray.plugins.operator_names()

    counts = cosmic_ray.counting.count_mutants(modules, operators)

    print('[Counts]')
    pprint.pprint(counts)
    print('\n[Total test runs]\n',
          sum(itertools.chain(
              *(d.values() for d in counts.values()))))


@dsc.command()
def handle_test_runners(config):
    """usage: {program} test-runners

List the available test-runner plugins.
"""
    print('\n'.join(cosmic_ray.plugins.test_runner_names()))
    return 0


@dsc.command()
def handle_operators(config):
    """usage: {program} operators

List the available operator plugins.
"""
    print('\n'.join(cosmic_ray.plugins.operator_names()))
    return 0


@dsc.command()
def handle_worker(config):
    """usage: {program} worker [options] <module> <operator> <occurrence> <test-runner> [-- <test-args> ...]

Run a worker process which performs a single mutation and test run. Each
worker does a minimal, isolated chunk of work: it mutates the <occurence>-th
instance of <operator> in <module>, runs the test suite defined by
<test-runner> and <test-args>, prints the results, and exits.

Normally you won't run this directly. Rather, it will be launched by celery
worker tasks.

options:
  --no-local-import   Disallow importing module from the current directory
  --keep-stdout       Do not squelch stdout
"""
    if not config['--no-local-import']:
        sys.path.insert(0, '')

    operator = cosmic_ray.plugins.get_operator(config['<operator>'])
    test_runner = cosmic_ray.plugins.get_test_runner(
        config['<test-runner>'],
        config['<test-args>'])

    with open(os.devnull, 'w') as devnull,\
        redirect_stdout(sys.stdout if config['--keep-stdout'] else devnull):
        work_record = cosmic_ray.worker.worker(
            config['<module>'],
            operator,
            int(config['<occurrence>']),
            test_runner)

    sys.stdout.write(
        json.dumps(work_record))


DOC_TEMPLATE = """{program}

Usage: {program} [options] <command> [<args> ...]

Options:
  -h --help     Show this screen.
  -v --verbose  Use verbose logging

Available commands:
  {available_commands}

See '{program} help <command>' for help on specific commands.
"""


def common_option_handler(config):
    if config['--verbose']:
        logging.basicConfig(level=logging.INFO)


def main(argv=None):
    dsc.main(
        'cosmic-ray',
        'cosmic-ray v.2',
        argv=argv,
        doc_template=DOC_TEMPLATE,
        common_option_handler=common_option_handler)


if __name__ == '__main__':
    main()



from pathlib import Path

from cosmic_ray.modules import find_modules, fixup_module_name
from path_utils import DATA_DIR, excursion, extend_path


def test_small_directory_tree():
    datadir = DATA_DIR
    paths = (('a', '__init__.py'),
             ('a', 'b.py'),
             ('a', 'py.py'),
             ('a', 'c', '__init__.py'),
             ('a', 'c', 'd.py'))
    expected = sorted(datadir / Path(*path) for path in paths)
    with extend_path(datadir):
        results = sorted(
            Path(m.__file__)
            for m in find_modules('a'))
    assert expected == results


def test_finding_modules_via_dir_name():
    datadir = DATA_DIR
    paths = (('a', 'c', '__init__.py'),
             ('a', 'c', 'd.py'))
    expected = sorted(datadir / Path(*path) for path in paths)
    with extend_path(datadir), excursion(datadir):
        module_name = fixup_module_name('a/c')
        results = sorted(
            Path(m.__file__)
            for m in find_modules(module_name))
    assert expected == results


def test_finding_modules_via_dir_name_and_filename_ending_in_py():
    datadir = DATA_DIR
    paths = (('a', 'c', 'd.py'),)
    expected = sorted(datadir / Path(*path) for path in paths)
    with extend_path(datadir), excursion(datadir):
        module_name = fixup_module_name('a/c/d.py')
        results = sorted(
            Path(m.__file__)
            for m in find_modules(module_name))
    assert expected == results


def test_finding_module_py_dot_py_using_dots():
    datadir = DATA_DIR
    paths = (('a', 'py.py'),)
    expected = sorted(datadir / Path(*path) for path in paths)
    with extend_path(datadir), excursion(datadir):
        module_name = fixup_module_name('a.py')
        results = sorted(
            Path(m.__file__)
            for m in find_modules(module_name))
    # a.py is not a path so we load a/py.py
    assert expected == results


def test_finding_modules_py_dot_py_using_slashes():
    datadir = DATA_DIR
    with extend_path(datadir):
        results = sorted(
            Path(m.__file__)
            for m in find_modules('a/py'))
    # a/py isn't a directory so nothing is loaded
    assert [] == results


def test_finding_modules_py_dot_py_using_slashes_with_full_filename():
    datadir = DATA_DIR
    paths = (('a', 'py.py'),)
    expected = sorted(datadir / Path(*path) for path in paths)
    with extend_path(datadir), excursion(datadir):
        module_name = fixup_module_name('a/py.py')
        results = sorted(
            Path(m.__file__)
            for m
            in find_modules(module_name))
    # a/py.py is a module and it is loaded
    assert expected == results



"""Celery specific details for routing work requests to Cosmic Ray workers."""
import celery
from celery.utils.log import get_logger
import itertools
import json
import subprocess

from .celery import app
from ..worker import WorkerOutcome
from ..work_record import WorkRecord

LOG = get_logger(__name__)


@app.task(name='cosmic_ray.tasks.worker')
def worker_task(work_record,
                test_runner,
                test_args,
                timeout):
    """The celery task which performs a single mutation and runs a test suite.

    This runs `cosmic-ray worker` in a subprocess and returns the results.

    Returns: An updated WorkRecord
    """
    # The work_record param may come as just a dict (e.g. if it arrives over
    # celery), so we reconstruct a WorkRecord to make it easier to work with.
    work_record = WorkRecord(work_record)

    command = list(itertools.chain(
        ('cosmic-ray',
         'worker',
         work_record.module,
         work_record.operator,
         str(work_record.occurrence),
         test_runner,
         '--',),
        test_args))

    LOG.info('executing: %s', command)

    proc = subprocess.Popen(command,
                            stdout=subprocess.PIPE,
                            universal_newlines=True)
    try:
        outs, _ = proc.communicate(input=None, timeout=timeout)
        result = json.loads(outs)
        work_record.update({
            k: v
            for k, v
            in result.items()
            if v is not None
        })
    except subprocess.TimeoutExpired as e:
        work_record.worker_outcome = WorkerOutcome.TIMEOUT
        work_record.data = e.timeout
        proc.kill()
    except json.JSONDecodeError as e:
        work_record.worker_outcome = WorkerOutcome.EXCEPTION
        work_record.data = e

    work_record.command_line = command
    return work_record


def execute_work_records(test_runner,
                         test_args,
                         timeout,
                         work_records):
    """Execute a suite of tests for a given set of work items.

    Args:
      test_runner: The `TestRunner` instance to use for running tests.
      test_args: The list of arguments to pass to the test runner.
      timeout: The max length of time to let a test run before it's killed.
      work_records: An iterable of `work_db.WorkItem`s.

    Returns: An iterable of WorkRecords.
    """
    return celery.group(
        worker_task.delay(work_record,
                          test_runner,
                          test_args,
                          timeout)
        for work_record in work_records)



import cosmic_ray.tasks.celery
import cosmic_ray.tasks.worker
from cosmic_ray.work_record import WorkRecord

# TODO: These should be put into plugins. Callers of execute() should pass an
# executor.


def local_executor(test_runner, test_args, timeout, pending_work):
    for work_record in pending_work:
        yield cosmic_ray.tasks.worker.worker_task(
            work_record,
            test_runner,
            test_args,
            timeout)


class CeleryExecutor:
    def __init__(self, purge_queue=True):
        self.purge_queue = purge_queue

    def __call__(self, test_runner, test_args, timeout, pending_work):
        try:
            results = cosmic_ray.tasks.worker.execute_work_records(
                test_runner,
                test_args,
                timeout,
                pending_work)

            for r in results:
                yield WorkRecord(r.get())
        finally:
            if self.purge_queue:
                cosmic_ray.tasks.celery.app.control.purge()


def execute(work_db, dist=True):
    """Execute any pending work in `work_db`, recording the results.

    This looks for any work in `work_db` which has no results, schedules to be
    executed, and records any results that arrive.

    If `dist` is `True` then this uses Celery to distribute tasks to remote
    workers; of course you need to make sure that these are running if you want
    tests to actually run! If `dist` is `False` then all tests will be run
    locally.
    """
    test_runner, test_args, timeout = work_db.get_work_parameters()
    executor = CeleryExecutor() if dist else local_executor
    work_records = executor(test_runner,
                            test_args,
                            timeout,
                            work_db.pending_work)
    for work_record in work_records:
        work_db.update_work_record(work_record)



from cosmic_ray.work_record import make_record
from hypothesis import given, assume
import hypothesis.strategies as ST
import pytest
import string

attributes = ST.text(alphabet=string.ascii_letters, min_size=1, max_size=10)


@given(attributes)
def test_empty_constructor_has_no_attributes(attr):
    r = make_record('Rec')()
    with pytest.raises(AttributeError):
        getattr(r, attr)


def test_empty_constructor_has_no_keys():
    r = make_record('Rec')()
    assert len(r) == 0


@given(ST.sets(attributes))
def test_default_attr_value_is_none(attrs):
    r = make_record('Rec', attrs)()
    for attr in attrs:
        assert getattr(r, attr) is None
    assert len(r) == len(attrs)


@given(attributes, ST.text())
def test_can_assign_to_field_attributes(attr, value):
    r = make_record('Rec', [attr])()
    setattr(r, attr, value)
    assert getattr(r, attr) == value


@given(attributes, attributes)
def test_cannot_assign_to_non_field_attributes(attr, nonattr):
    assume(attr != nonattr)
    r = make_record('Rec', [attr])()
    with pytest.raises(AttributeError):
        r.bar = 'llama'


@given(attributes)
def test_can_index_fields(attr):
    r = make_record('Rec', [attr])()
    r[attr] = 42
    assert r[attr] == 42


@given(attributes, attributes)
def test_cannot_index_non_fields(attr, nonattr):
    assume(attr != nonattr)
    r = make_record('Rec', [attr])()
    with pytest.raises(KeyError):
        r[nonattr] = 42
    with pytest.raises(KeyError):
        r[nonattr]


@given(attributes)
def test_can_update_fields(attr):
    r = make_record('Rec', [attr])()
    r.update({attr: 42})
    assert r[attr] == 42


@given(attributes, attributes)
def test_cannot_update_non_fields(attr, nonattr):
    assume(attr != nonattr)
    r = make_record('Rec', [attr])()
    with pytest.raises(KeyError):
        r.update({nonattr: 42})


@given(attributes)
def test_cannot_delete_fields(attr):
    r = make_record('Rec', [attr])()
    with pytest.raises(KeyError):
        del r[attr]


@given(ST.sets(attributes))
def test_can_construct_from_dict(attrs):
    make_record('Rec', attrs)({a: 42 for a in attrs})


@given(attributes, attributes)
def test_cannot_construct_from_dict_with_non_fields(attr, nonattr):
    assume(attr != nonattr)
    R = make_record('Rec', [attr])
    with pytest.raises(KeyError):
        R({nonattr: 3})



"""Central location for celery-specific stuff."""

from celery import Celery

app = Celery(
    'cosmic-ray',
    broker='amqp://',
    backend='amqp://')

app.conf.CELERY_ACCEPT_CONTENT = ['json']
app.conf.CELERY_TASK_SERIALIZER = 'json'
app.conf.CELERY_RESULT_SERIALIZER = 'json'

# This will remove all pending work from the queue. We need to do this when we
# shut down during exec:
#
#     cosmic_ray.celery.app.control.purge()



import ast


class PrintVisitor(ast.NodeVisitor):
    """A `NodeVisitor` which simply prints information about the nodes it
    visits.
    """

    def __init__(self):
        self.indent = ''

    def generic_visit(self, node):
        """Print `node` and manage an indentation state."""
        print(self.indent, repr(node))
        self.indent += '    '
        super().generic_visit(node)  # pylint:disable=missing-super-argument
        self.indent = self.indent[:-4]

    def visit_Num(self, node):  # noqa
        print('a number:', node)


def dump_mod():
    import mod
    with open(mod.__file__, 'rt') as module_file:
        nodes = ast.parse(module_file.read())
    PrintVisitor().visit(nodes)



import logging
import uuid

import cosmic_ray.modules
from cosmic_ray.work_record import WorkRecord

LOG = logging.getLogger()


def init(modules,
         work_db,
         test_runner,
         test_args,
         timeout):
    """Clear and initialize a work-db with work items.

    Any existing data in the work-db will be cleared and replaced with entirely
    new work orders. In particular, this means that any results in the db are
    removed.

    """
    operators = cosmic_ray.plugins.operator_names()
    counts = cosmic_ray.counting.count_mutants(modules, operators)
    work_db.set_work_parameters(
        test_runner=test_runner,
        test_args=test_args,
        timeout=timeout)

    work_db.clear_work_records()

    work_db.add_work_records(
        WorkRecord(
            job_id=uuid.uuid4().hex,
            module=module.__name__,
            operator=opname,
            occurrence=occurrence)
        for module, ops in counts.items()
        for opname, count in ops.items()
        for occurrence in range(count))



"""Facilities for generating ASTs from modules."""

import ast
import inspect
import logging

LOG = logging.getLogger()

# TODO: This is where we can do different things for different kinds of
# modules. Right now we only really handle normal source-code, text-file
# modules.


def get_ast(module):
    """Generate an AST from a module object.

    This will be the AST for the contents of the module.

    Raises:
        OSError: If the source code for `module` can't be found.
        TypeError: If the source file for `module` can't be found.
    """
    try:
        source_file = inspect.getsourcefile(module)
    except TypeError:
        LOG.error("Unable to get source file for object %s", module)
        raise

    try:
        source = inspect.getsource(module)
    except OSError:
        # workaround for empty __init__.py files, see
        # http://bugs.python.org/issue27578
        if source_file.endswith('__init__.py'):
            source = ''
        else:
            LOG.info('inspect.getsource() failed.'
                     'Attempting to read source directly: %s',
                     module)
            with open(source_file, mode='rt') as f:
                source = f.read()

    return ast.parse(source, source_file, 'exec')
